# CloudFormation does not currently support.
# 
# September 2021
# 
# https://github.com/bridgecrewio/checkov/blob/master/checkov/terraform/checks/resource/aws/EKSControlPlaneLogging.py
# 
# https://github.com/aws/containers-roadmap/issues/242
# https://github.com/aws-cloudformation/cloudformation-coverage-roadmap/issues/208
